#==============================================================================
#
#  $Id: composer.py,v 1.14 2001/03/30 01:37:45 mike Exp $
#
"""
   This module houses all of the front end components for message composition.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: composer.py,v $
#  Revision 1.14  2001/03/30 01:37:45  mike
#  Separated "quit" from "terminate" so that we don't give the "contents
#  changed" warning after sending a message.
#
#  Revision 1.13  2001/03/18 18:51:41  mike
#  Added an "are you sure" query before quitting with changes made to the
#  body.
#
#  Revision 1.12  2001/03/17 22:30:59  mike
#  Exceptions thrown while sending a message no longer take down the composer.
#
#  Revision 1.11  2000/07/27 23:35:41  mike
#  Fixed bug causing "attach file" to blow up if the file wasn't found.
#
#  Revision 1.10  2000/07/16 23:24:41  mike
#  Added support for address groups to composer.
#
#  Revision 1.9  2000/06/25 21:36:37  mike
#  Fixed attachment name construction - again.
#
#  Revision 1.8  2000/05/28 19:53:51  mike
#  Added the ability to save messages to the "Drafts" folder.
#
#  Revision 1.7  2000/05/25 01:26:47  mike
#  Added support for adding attachments.
#
#  Revision 1.6  2000/05/13 00:05:26  mike
#  Added hack to cleanup editor-created backup of temporary file used for
#  composition.
#
#  Revision 1.5  2000/04/30 21:07:22  mike
#  Added initial address book window.
#
#  Revision 1.4  2000/04/27 23:16:16  mike
#  Took out the "send" button, added check for recipients.
#
#  Revision 1.3  2000/04/17 12:33:50  mike
#  Added status indicators.
#
#  Revision 1.2  2000/04/02 22:38:01  mike
#  Added code to cleanup recipient lines, and add the subject and body before
#  sending the message.
#
#  Revision 1.1.1.1  2000/04/02 19:09:03  mike
#  Spug Mailer and libraries.
#
#
#==============================================================================


import tempfile, os, string, re, base64

try:
   from cStringIO import StringIO
except:
   from StringIO import StringIO

from spug.curses.KeyEvent import KeyEvent
from spug.curses.ModalWindow import ModalWindow
from spug.curses.Label import Label
from spug.curses.Entryfield import Entryfield
from spug.curses.Listbox import Listbox
from spug.curses.SimpleTextViewer import SimpleTextViewer
from spug.curses.Button import Button

from addrbk import GroupEntry
from message import MessageHeaders, Message
from senders import SendError

import main

class HeaderEntryfield(Entryfield):
   
   def __init__(self, x0, y0, width, maxEntry, content, onEnter):
      Entryfield.__init__(self, x0, y0, width, maxEntry, content)
      self.onEnter = onEnter
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent) and evt.code == 10:
         self.onEnter()
         return 1
      else:
         return Entryfield.processEvent(self, evt)

class HeaderItem:
   
   def __init__(self, type, content, memberOf = None, group = None):
      self.type = type
      self.content = content
      self.memberOf = memberOf
      self.group = group

class HeaderList(Listbox):
   
   def __init__(self, parent, x0, y0, width, height, items, parent):
      Listbox.__init__(self, parent, x0, y0, width, height, items)
      self.parent = parent
   
   def _itemAsString(self, item):
      if item.memberOf:
         return '%-8s    %s' % (item.type + ':', item.content)
      else:
         return '%-8s %s' % (item.type + ':', item.content)

   def __deleteItem(self, itemToDelete, index):
      # and delete it
      self.delete(index)
      
      # if this is a group item, delete all of the members
      if itemToDelete.group:
         items = self.getItems()
         items = map(None, items, range(len(items)))
         items.reverse()
         for item, i in items:
            if item.memberOf is itemToDelete:
               self.delete(i)
      

   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == self.KEY_DC:
            # delete the header line
            index = self.selectionIndex()
            if index != -1:            
               self.__deleteItem(self.selection(), index)
               
            return 1
         elif evt.code == 10:
            item = self.selection()
            self.parent.setHeaderField(item)
            
            # again, delete the header line
            index = self.selectionIndex()
            self.__deleteItem(item, index)
            
      return Listbox.processEvent(self, evt)

class AttachmentItem:
   
   def __init__(self, file, name, attachment):
      self.file = file
      self.name = name
      self.attachment = attachment

class AttachmentList(Listbox):
   
   def __init__(self, parent, x, y, width, height, items):
      Listbox.__init__(self, parent, x, y, width, height, items)
   
   def _itemAsString(self, item):
      return '%s "%s"' % (item.file, item.name)

class BodyViewer(SimpleTextViewer):
   
   def __init__(self, parent, x0, y0, width, height, text):
      SimpleTextViewer.__init__(self, parent, x0, y0, width, height, text)
      self.text = text
      self.parent = parent
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == 10:
            temp = tempfile.mktemp()
            open(temp, 'w').write(self.text)
            main.mainWin.launchApp('echo \033c; stty sane; jed %s' % 
                                    temp, 
                                   self.parent
                                   )
            self.text = open(temp).read()
            os.remove(temp)
            
            # remove the backup file if the editor generated one
            if os.path.exists(temp + '~'):
               os.remove(temp + '~')
            
            self.setText(self.text)
            self.paint()
            return 1
            
      return SimpleTextViewer.processEvent(self, evt)      

class Composer(ModalWindow):
   """
      Message composition window.
      
      Public-vars:
         messageWasSaved::
            [@mailbx.MailboxMessageInfo or None] if the user exited the
	    composer with the "save to drafts" option, this is the message
	    info instance, otherwise it is None.

   """
   
   __headerTypes = [ 'to', 'cc', 'bcc' ]
   __addrSepRx = re.compile(r',\s*')
   
   def loadHeaders(self):
      """
         Loads the headers from the message into the listbox.
      """
      for hdr, contents in self.msg.headers:
         if hdr in ('to', 'cc', 'bcc'):
            for line in string.split(contents, '\n'):
               line = string.strip(line)
               self.headerLst.append(HeaderItem(hdr, line))

   def __addAttachments(self):
      """
         Loads message attachments into the listbox.
      """
      for att in self.msg.attachments:
      
         # check to see if there is a filename in the content disposition
         name = ''
         if att.headers.has_key('content-disposition'):
            try:
               type, parms = att.headers.parseParms('content-disposition')
               if type == 'attachment':
                  if parms.has_key('filename'):
                     name = parms['filename']
                  elif parms.has_key('name'):
                     name = parms['name']
            except TypeError:
               pass
         
         # check to see if there was a comment
         if att.headers.has_key('content-description'):
            comment = att.headers['content-description']
         else:
            comment = name
         
         # if we still don't have a name, just use content type
         if not name:
            name = att.getContentType()[0]
         
         self.attachments.append(AttachmentItem(name, comment, att))
   
   def __init__(self, x0, y0, width, height, message):
      ModalWindow.__init__(self, x0, y0, width, height)
      self.headerTypeLst = Listbox(self, 0, 0, 8, 1, self.__headerTypes)
      self.headerFld = HeaderEntryfield(self, 8, 0, width - 8, 80, '')
      self.headerFld.onEnter = self.addHeader
      self.headerLst = HeaderList(self, 0, 1, width, 4, [], self)
      self.subjectLbl = Label(self, 0, 5, 'Subject:')
      self.subjectFld = Entryfield(self, 8, 5, width - 8, 80,
                                   message.headers['subject']
                                   )
      self.body = BodyViewer(self, 0, 6, width, height - 9, 
                             message.body
                             )
      self.attachmentsLbl = Label(self, 0, height - 3, 'Attachments:')
      self.attachments = AttachmentList(self, 0, height - 2, width, 2, 
                                        []
                                        )
      self.msg = message      
      self.loadHeaders()
      self.__addAttachments()
      
      self.messageWasSaved = None
      
      self.refresh()      

   def terminate(self):
      """
         Overrides @spug.curses.ModalWindow.ModalWindow.terminate(), cleans
         up cyclic references.
      """
      
      del self.headerFld, self.body, self.headerLst
      ModalWindow.terminate(self)

   def quit(self):
      """
         Called when the quit option is pressed.
      """
      # make sure that there is nothing worth saving
      if self.msg.body != self.body.text and \
         not main.mainWin.queryYesNo('Discard changes?'):
         return
      
      self.terminate()

   def addGroup(self, group):
      groupItem = \
         HeaderItem(self.headerTypeLst.selection(), group.fullName + ' -', 
                    None,
                    group
                    )
      self.headerLst.append(groupItem)
      
      for member in group.members:
         member = main.addressBook[member]
         self.headerLst.append(HeaderItem(self.headerTypeLst.selection(),
                                          member.asAddress(),
                                          groupItem
                                          )
                               )
         
   
   def addHeader(self):
      # get the header content
      content = self.headerFld.getText()
      
      # check to see if it is an address book key
      strippedContent = string.strip(content)
      if main.addressBook.has_key(strippedContent):
         entry = main.addressBook[strippedContent]
         
         # if this is a group entry, we process it very differently
         if isinstance(entry, GroupEntry):
            self.addGroup(entry)
            self.headerFld.setText('')
            return
         else:
            content = entry.asAddress()
      
      self.headerLst.append(HeaderItem(self.headerTypeLst.selection(),
                                       content
                                       )
                            )
      self.headerLst.refresh()
      self.headerFld.setText('')

   def setHeaderField(self, headerItem):
      """
         Sets the header field controls from the given header item.
     
         parms:
            headerItem::
               [@HeaderItem] the header info to enter
      """
      self.headerTypeLst.setSelection(headerItem.type)
      if headerItem.group:
         self.headerFld.setText(headerItem.group.nickname)
      else:
         self.headerFld.setText(headerItem.content)
      self.setFocus(self.headerFld)
   
   def __fixHeaders(self):
      """
         "Fixes" the headers in self.msg by removing all "to", "cc", and
	 "bcc" fields and replacing them with the contents of the header
	 list.
      """
      # delete all headers in question
      for type in self.__headerTypes:
         
         # keep deleting all headers of that type till there are none left
         while 1:
            try:
               del self.msg.headers[type]
            except KeyError:
               break

      # if the header content entryfield is not empty, add it to the list
      if self.headerFld.getText():
         self.addHeader()
      
      # consolidate all headers in the main line into the appropriate groups
      
      groups = {}
      lastGroup = ''
      for item in self.headerLst.getItems():
         if item.group:
            # if this is a group entry, hold onto it so that we can append
            # to the next non-group entry
            lastGroup = item.content + ' '
            continue
         else:
            
            # convert the content string to a list of address strings
            val = self.__addrSepRx.split(item.content)
            
            # if this was preceeded by a group entry, prepend the group
            # name to the first address string
            if lastGroup and val:
               val[0] = lastGroup + val[0]
               lastGroup = ''

         # add the list of addresses to the group of addresses under that
         # header type
         if groups.has_key(item.type):
            groups[item.type] = groups[item.type] + val
         else:
            groups[item.type] = val

      # now we reformat the headers and add them to the message
      
      for hdr in self.__headerTypes:
         if groups.has_key(hdr):
            
            # create a new string for the current line and the total value
            line = val = ''
            
            for addr in groups[hdr]:
               
               # ignore any empty addresses
               if not addr:
                  continue
               
               # concatenate the current address onto the line, breaking
               # the line and starting a new one if it gets too wide.
               if not line:
                  line = line + addr
               elif len(line) + len(addr) + 2 >= 72:
                  if val:
                     val = '%s,\n\t%s' % (val, line)
                  else:
                     val = line
                  line = addr
               else:
                  line = '%s, %s' % (line, addr)
            
            # tack on the last line
            if val:
               val = '%s,\n\t%s' % (val, line)
            else:
               val = line
            
            # add the header 
            self.msg.headers.addHeader(hdr, val)
      
      # keep the subject header at the end of the list
      self.msg.headers.moveToEnd('subject')

   def __checkHeaders(self):
      """
         Checks to make sure that the header list is ok.  Reports errors
         on the status line.
         
         At this time, just checks that there's something there.  Returns
         true if everything is ok, false if not.
      """
      if string.strip(self.headerFld.getText()):
         return 1
      
      if self.headerLst.getItems():
         return 1
      
      main.mainWin.showStatus('You must specify a recipient')
      self.beep()

   def __storeFields(self):
      """
         Stores all fields in the composer to the message.
      """
      self.__fixHeaders()
      self.msg.body = self.body.text
      self.msg.headers['subject'] = self.subjectFld.getText()      
   
   def sendMsg(self):   
      # make sure that the recipients are ok
      if not self.__checkHeaders():
         return
      
      self.__storeFields()
      self.msg.fixAttachments()
      main.mainWin.showStatus('Sending message...')
      
      # try to send the message, catch all exceptions
      try:
         main.custom.sendMessage(self.msg)
      except SendError, ex:
         main.mainWin.showStatus('Error: %s' % ex[0])
         return
      except Exception, ex:
         import traceback, sys, string
         info = apply(traceback.format_exception, sys.exc_info())
         main.mainWin.showText(string.join(info, ''), self)
         return
         
      main.postOffice.getMailbox('Sent').addMessage(self.msg)
      main.mainWin.showStatus('Message sent and copied to "Sent" folder')
      self.terminate()

   def saveMessage(self):
      self.__storeFields()
      self.messageWasSaved = \
         main.postOffice.getMailbox('Drafts').addMessage(self.msg)
      main.mainWin.showStatus('Message saved to "Drafts"')

      self.terminate()

   def attachFile(self):
      # get the filename from the user and make sure that it exists
      filename = main.mainWin.queryUser('Filename:')
      if not os.path.exists(filename):
         main.mainWin.showStatus('"%s" not found' % filename)
         return
      
      # get the comment from the user
      comment = main.mainWin.queryUser('Comment:')
      
      # read and encode the file
      out = StringIO()
      main.mainWin.showStatus('Reading %s...' % filename)
      base64.encode(open(filename, 'rb'), out)
      
      # get the specific filename for the header
      baseName, specificFilename = os.path.split(filename)
      
      headers = MessageHeaders()
      headers['content-type'] = 'text/plain; charset="us-ascii"'
      headers['content-transfer-encoding'] = 'base64'
      headers['content-disposition'] = \
         'attachment; filename="%s"' % specificFilename
      headers['content-description'] = comment
      msg = Message(headers, out.getvalue(), [])
      
      self.attachments.append(AttachmentItem(filename, comment, msg))
      self.msg.attachments.append(msg)
      main.mainWin.showStatus('Attachment added')

   def deleteAttachment(self):
      """
         Removes the selected attachment from the message.  Since this
         can have painful consequences, prompts the user with an "are
         you sure?"
      """
      i = self.attachments.selectionIndex()
      if i == -1: 
         self.beep()
         main.mainWin.showStatus('You must select an attachment to delete')

      if main.mainWin.queryYesNo('Delete selected attachment?'):
         self.attachments.delete(i)
         del self.msg.attachments[i]
         
   def processEvent(self, evt):
      # first let the kids at it
      if ModalWindow.processEvent(self, evt):
         return 1

      if isinstance(evt, KeyEvent):
         if evt.code == 24: # Ctrl-X
            main.menu( [ (115, 'Send', self.sendMsg),
                         (118, 'saVe to drafts', self.saveMessage),
                         (97, 'Attach file', self.attachFile),
                         (100, 'Delete attachment', self.deleteAttachment),
                         (113, 'Quit', self.quit) 
                        ], 
                      self
                      )
            return 1
      
      return 0
